#include "lsample.h"

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
void  seq_imp (int nsmpls, int nped, int numloc, int *nall,
               double **Qq, int ***y,
               int schnell_flag,
               double *trt_ctrd,
               double var_err)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

     /* Written by: AWG, April, 2001
        Modified to include linked traits, by MAJ, May, 2001
        Purpose:
        Previously S_M was initalized by assuming unlinked loci.
        Sequential imputation is being used to come up with a better
        starting value for S_M.

        Method:
        Using sequential imputation, generate S_j^* and associated weights
        w_j (i.e. just single locus peeling). Then S^* and w^* are easily
        obtained. Choose an S^* as a starting value with prob \frac{w^*}{W}
        where W = sum w^*

        Variables:
        nsmpls - number of S_M's in which to impute via sequential imputation
        nped - number of individuals in the pedigree
        numloc - number of loci (markers + trait)
        nall  - number of observed alleles segregating at loci
        (marker + trait loci)
        Qq  - frequencies of observed marker and trait locus alleles
        y - observed marker and trait data
        trt_ctrd - centered quantitative trait values (ym)
        var_err - residual variance (ve)
      */

{
   struct seq_smpls
   {
      double weight;
      int **xx0, **xx1;                     /* array of maternal and paternal meiosis */
   };                                       /* indicators respectively */
   typedef struct seq_smpls Seq_smpls;

   int   j, jj, sx, k, kk;
   int   mxjj;                              /* max index associated with max normalized
                                               weight */
   int   rnloc;                             /* randomly selected locus */
   double *xecopy[2];                       /* copy of xecomo[][0][]; but malloced
                                               [2][numloc] */
   double mxwght;                           /* max normalized weight */
   double *seq_imp_w;                       /* ptr to array of nsmpls weights */

   Seq_smpls *samples;                      /* ptr to array structure with fields
                                               xx0, xx1 and weights */

   /* Setup: allocation of samples structure */
   samples = malloc (nsmpls * sizeof (Seq_smpls));
   alloc_chk (WHERE, samples);
   for (jj = 0; jj < nsmpls; jj++)
   {
      samples[jj].xx0 = malloc ((numloc) * sizeof (int *));
      alloc_chk (WHERE, samples[jj].xx0);
      samples[jj].xx1 = malloc ((numloc) * sizeof (int *));
      alloc_chk (WHERE, samples[jj].xx1);
      samples[jj].xx0[0] = malloc ((nped + 1) * numloc * sizeof (int));
      alloc_chk (WHERE, samples[jj].xx0[0]);
      for (j = 1; j < numloc; ++j)
         samples[jj].xx0[j] = samples[jj].xx0[j - 1] + (nped + 1);
      samples[jj].xx1[0] = malloc ((nped + 1) * numloc * sizeof (int));
      alloc_chk (WHERE, samples[jj].xx1[0]);
      for (j = 1; j < numloc; ++j)
         samples[jj].xx1[j] = samples[jj].xx1[j - 1] + (nped + 1);
   }

   /* Setup: allocate structures needed for the L-sampler  */
   if (!schnell_flag)
   {
      allocate_loci (numloc, nall, Qq);
   }

   /* and allocate array xecopy[2][] for copy of xecomo[][0][] */
   /* copy xecomo[][0][] to the array */

   for (sx = FEM; sx <= MAL; sx++)
   {
      xecopy[sx] = malloc ((numloc) * sizeof (double));
      alloc_chk (WHERE, xecopy[sx]);
      for (kk = 0; kk < numloc; kk++)
         xecopy[sx][kk] = xecomo[sx][0][kk];
   }
   if (schnell_flag)
   {
      /* For Schnell */
      n_markers = numloc - 1;           /* set globals    */
      n_tloci = 1;
   }
   else
   {
      /* For Autozyg */
      n_markers = numloc;
      n_tloci = 0;
   }

   fputc ('\n', stdout);
   locus_peel_init (0, 1, y);
   peel_trace = 0;

   for (jj = 0; jj < nsmpls; jj++)      /* draw nsmpls via seq. imp. */
   {
      /* Begin with loci on seperate linkage groups
         i.e. locusrec re-initialized */
      for (sx = FEM; sx <= MAL; sx++)
         for (kk = 0; kk < numloc; kk++)
            xecomo[sx][0][kk] = 0.5;

      /* specify map for L sampler */
      set_positions (numloc - 1);

      set_transmission (0, 0, 0);

      /* randomly select a marker or trait locus upon which to commence
         sequential imputation */
      rnloc = intran (0, numloc - 1);

      seq_imp_w = calloc (numloc, sizeof (double));
      alloc_chk (WHERE, seq_imp_w);

      /* generate S's for unlinked trait locus **** removed **** */

      /* generate S's (as if unlinked) for randomly selected locus */
      fflush (stdout);
      seq_imp_w[rnloc] = locus_mjrpeel (trt_ctrd, var_err, 1.0,
                                        1, loci + rnloc);
      fflush (stdout);
      if (seq_imp_w[rnloc] == -DBL_MAX)
         give_up_where (WHERE, ERA_wgt, rnloc);

      /* Backwards sequential imputation */
      for (kk = rnloc - 1; kk >= 0; kk--)
      {
         for (sx = FEM; sx <= MAL; sx++)
            xecomo[sx][0][kk] = xecopy[sx][kk];
         set_positions (numloc - 1);
         fflush (stdout);
         set_transmission (loci, kk, numloc);
         seq_imp_w[kk] = locus_mjrpeel (trt_ctrd, var_err, 1.0, 1, loci + kk);
         fflush (stdout);
         if (seq_imp_w[kk] == -DBL_MAX)
            give_up_where (WHERE, ERA_wgt, kk);
      }

      /* Forwards sequential imputation */
      for (kk = rnloc + 1; kk < numloc; kk++)
      {
         for (sx = FEM; sx <= MAL; sx++)
            xecomo[sx][0][kk - 1] = xecopy[sx][kk - 1];
         set_positions (numloc - 1);
         fflush (stdout);
         set_transmission (loci, kk, numloc);
         seq_imp_w[kk] = locus_mjrpeel (trt_ctrd, var_err, 1.0, 1, loci + kk);
         fflush (stdout);
         if (seq_imp_w[kk] == -DBL_MAX)
            give_up_where (WHERE, ERA_wgt, kk);
      }

      /* add seq_imp_w in preparation for the calculation of the weight for i
         S_M^*  */
      samples[jj].weight = 0.0;
      for (kk = 0; kk < numloc; kk++)
         samples[jj].weight += seq_imp_w[kk];

      /* copy contents of xx0 and xx1 into samples structure */
      for (k = 0; k < numloc; k++)
         for (kk = 0; kk <= nped; kk++)
         {
            samples[jj].xx0[k][kk] = xx0[k][kk];
            samples[jj].xx1[k][kk] = xx1[k][kk];
         }
      if (seq_imp_w)
         free (seq_imp_w);
   }                                    /* end sequential imputation loop */

   mxwght = samples[0].weight;
   jj = 0;

   for (jj = 0; jj < nsmpls; jj++)
      if (mxwght <= samples[jj].weight)
      {
         mxwght = samples[jj].weight;
         mxjj = jj;
      }

   /* assign contents of samples to global vars xx0 and xx1 */

   for (k = 0; k < numloc; k++)
      for (kk = 0; kk <= nped; kk++)
      {
         xx0[k][kk] = samples[mxjj].xx0[k][kk];
         xx1[k][kk] = samples[mxjj].xx1[k][kk];
      }

   /* free structures */
   for (jj = 0; jj < nsmpls; jj++)
   {
      free (samples[jj].xx0[0]);
      free (samples[jj].xx1[0]);
   }
   for (jj = 0; jj < nsmpls; jj++)
   {
      free (samples[jj].xx0);
      free (samples[jj].xx1);
   }
   free (samples);

   for (sx = FEM; sx <= MAL; sx++)
      free (xecopy[sx]);

   return;

}
