
/*
 * proc_auto() has as its input the results of parsing the parameter
 * statements by proc_pars() for the Autozyg programs.  It checks this
 * data for consistency and completeness and, if there are no errors,
 * builds data structures used by lm_auto, lm_lods or ibddrop.
 *
 * this code is derived from proc_drop
 * one chromosome only is selected
 * there must be at least one marker
 * one trait only (linked or unlinked for lm_auto or ibddrop, positioned
 * along the chromosome for lm_lods) is expected.
 * any combination of sex-not-specified maps or sex-specific maps
 * may be supplied
 *                                                        maj      5/2000 
 *                                                        maj rev  1/2002
 *
 * in the comments below: a - lm_auto, l - lm_lods, i - ibddrop
 */

int   proc_auto (void)
{
   int   lnkd, m_no, entry, sex;
   int   sxsp_map = FALSE;
   SIM_PTR p_s;
   DIST_PTR p_d;
   UNLNKD_PTR p_u;
   TLIST_PTR p_t;
   VMAP_PTR p_v;
   PRNTMAP_PTR p_m;
   PRNTLOC_PTR p_l;

   if (!test_atz_sel ())                /* checks for select statement       */
      if (proc_err ())
         return FALSE;
   if (n_err)
      return proc_end ();

   entry = FALSE;                       /* selects one chromosome only       */
   p_s = nxt_sim_chrom (&entry);

   if (!test_mark_map (p_s, &sxsp_map)) /* checks marker map & trait map     */
      if (proc_err ())                  /* (sex-averaged & sex-specific)     */
         return FALSE;
   if (pmx->pcod == 'a' || pmx->pcod == 'i')
      if (!test_trait_map (p_s, &sxsp_map))
         if (proc_err ())
            return FALSE;
   if (pmx->pcod == 'l')
      if (!test_trait_vmap (p_s, &sxsp_map))
         if (proc_err ())
            return FALSE;
   if (n_err)                           /* if errors, quit                   */
      return proc_end ();

   lnkd = TRUE;                         /* adds mapped trait to trait list   */
   entry = FALSE;                       /* (a, i)                            */
   if (p_d = nxt_mapd_trait (&entry, FEM, p_s))
      if (!test_trait_list (lnkd, p_d->id))
         if (proc_err ())
            return FALSE;

   lnkd = FALSE;                        /* adds unlinked trait to trait list */
   entry = FALSE;                       /* (a, i)                            */
   if (p_u = nxt_unlnkd_trait (&entry))
      if (!test_trait_list (lnkd, p_u->id))
         if (proc_err ())
            return FALSE;

   lnkd = FALSE;                        /* adds trait with varying positions */
   if (p_v = nxt_trait_vmap (FEM, p_s)) /* to trait list (l)                 */
      if (!test_trait_list (lnkd, p_v->id))
         if (proc_err ())
            return FALSE;

   entry = FALSE;                       /* for ea marker, check allele freqs */
   if (pmx->pcod != 'i')
      while (m_no = nxt_mark (&entry, p_s))
         if (!test_mark_loci (p_s, m_no))
            if (proc_err ())
               return FALSE;

   entry = FALSE;                       /* for the trait, check allele freqs */
   if (pmx->pcod != 'i')
      if (p_t = nxt_trait (&entry))
         if (!test_trait_loci (p_t->id))
            if (proc_err ())
               return FALSE;

   if (n_err)
      return proc_end ();

   for (sex = FEM; sex <= AVG; ++sex)   /* if no errors, prints out params   */
   {                                    /* first assigns loci to sex-specific 
                                         * print maps         
                                         * for maps not present, defaults to 
                                         * sex-averaged,  then to F maps     */

      p_m = add_prntmap (p_s, sex);     /* puts markers into map             */

      if (p_s->no_mark == 1)
         add_only_mark_locus (p_m);
      else
      {
         entry = FALSE;
         while (p_d = nxt_mapd_mark (&entry, sex, p_s))
            add_mark_locus (p_m, p_d);
      }
      entry = FALSE;                    /* adds traits to map (a, i)         */
      if (p_d = nxt_mapd_trait (&entry, sex, p_s))
         add_trait_locus (p_m, p_d);

      entry = FALSE;                    /* finds distances between adj loci  */
      while (p_l = nxt_prntmap_locus (&entry, p_m))
         fix_locus_dist (p_m, p_l);

      if ((sxsp_map && (sex == FEM || sex == MAL))
          || (!sxsp_map && sex == AVG))
      {
         prnt_map (p_m, sex);           /* prints the map                    */

         entry = FALSE;                 /* checks trait distances (a, i)     */
         while (p_l = nxt_prntmap_locus (&entry, p_m))
            if (!test_trait_dist (p_m, p_l))
               if (proc_err ())
                  return FALSE;

         if (p_v = nxt_trait_vmap (sex, p_s))  /* prints the trait intervals */
            prnt_vmap (p_v, sex);
      }
   }

   if (pmx->pcod != 'i')
   {
      prnt_mark_freqs (p_s);            /* prints marker allele freqs        */

      prnt_trait_freqs (p_m);           /* prints trait allele freqs         */

      prnt_unlnkd_freqs ();
   }
   prnt_vtrait_freqs (p_s);

   fprintf (stdout, "\n");

   if (n_err)
      return proc_end ();

   NMarks = p_m->no_mark;               /* counts loci                       */
   NLnkd = p_m->no_traits;
   if (p_u)
      ++NUnlnkd;
   if (p_v)
      ++NVary;
   NTraits = NLnkd + NUnlnkd + NVary;

   NLoci = NMarks + NTraits;

   alloc_marek ();                      /* stores marker recomb fracts in
                                         * Marek[][]                         */
   mak_marek ();

   if (NLnkd || NUnlnkd)                /* stores trait recomb fract and pos */
      mak_tarek ();

   if (NVary)                           /* sets NTPos, the no. of trait pos  */
   {                                    /* (arrays are 1 greater)            */
      entry = FALSE;                    /* and recomb fracts in RecVal[][]   
                                         * and positions in IntVal[]         */

      NTExtPos = p_v->no_xtrn;          /* ADDED BY AWG for lm_bayes         */
      NTIntPos = p_v->no_prop;          /* END OF ADDED SECTION              */

      NTPos = 2 * p_v->no_xtrn + (NMarks - 1) * p_v->no_prop + 1;

      alloc_intval ();

      mak_intval (p_s);

      mak_recval (p_s);
   }

   entry = FALSE;                       /* for each locus, stores allele     */
   p_m = nxt_prntmap (&entry, FEM);     /* frequencies in Qq[][]             */
   entry = FALSE;
   if (pmx->pcod != 'i')
   {
      while (p_l = nxt_prntmap_locus (&entry, p_m))
         mak_allefrq (p_m, p_l, p_l->id);
      if (p_u)
         mak_allefrq (NULL, NULL, p_u->id);
      if (p_v)
         mak_allefrq (NULL, NULL, p_v->id);
   }

   if (pmx->pcod == 'a' || pmx->pcod == 'l')
      if (!test_markdat (p_s))          /* checks marker data                */
         if (proc_err ())
            return FALSE;

   if (!n_err)
   {
      mak_markdat (p_s);                /* stores and prints marker data     */
      prnt_markdat (p_s);
   }

   if (pmx->pcod == 'a' || pmx->pcod == 'i')
   {                                    /* checks proband gametes            */
      if (!test_probgam (p_s))          /* and locus window                  */
         if (proc_err ())
            return FALSE;
      if (!n_err)
      {                                 /* stores and prints proband gametes */
         mak_probgam ();
         prnt_probgam ();
      }
   }

   if (!test_atz_pars ())               /* checks computational parms        */
      if (proc_err ())
         return FALSE;

   if (!test_atz_inpt ())               /* checks input parmeters            */
      if (proc_err ())
         return FALSE;

   if (n_err)
      return proc_end ();

   prnt_atz_pars ();                    /* prints program parameters         */

   mak_atz_inpt ();                     /* set up stuff for reading          */

   return TRUE;
}

/* -------- test fncs:  test results of processing input statements -------- */

/* ----------------------- test functions for autozyg ---------------------- */

VMAP_PTR nxt_trait_vmap (int sx, SIM_PTR p_sim)
{                                       /* returns the pointer to the trait  */
   /*                                    * map for the selected chromosome   */
   /*                                    * and gender                        */
   int   i, j;
   VMAP_PTR p, p_vmap = NULL;

   for (i = 1; !p_vmap && i <= 4; ++i)
   {
      if (i == 1)                       /* checks for gender-specific map    */
         j = sx;                        /* present in this order             */
      else if (i == 2)
         j = AVG;
      else if (i == 3)
         j = FEM;
      else if (i == 4)
         j = MAL;

      if (p = first_trait_vmap_node[j])
      {
         while (p->p_next && (p->chrom_no < p_sim->chrom_no))
            p = p->p_next;
         if (p->chrom_no == p_sim->chrom_no)
            p_vmap = p;
      }
   }
   return p_vmap;
}

int   test_atz_sel (void)
{                                       /*checks select statements           */
   int   n;
   UNLNKD_PTR p;

   if (!first_sim_node)                 /*select ... markers .. required     */
      err_typ = ERA_SEL;

   else if (first_sim_node->p_next)     /*checks for only one chromosome     */
      err_typ = ERA_OOS;

   else
   {                                    /*checks for one mapped or           */
      n = first_sim_node->no_traits;    /*unlinked trait                     */

      if (p = first_unlnkd_node)
      {
         n++;
         while (p = p->p_next)
            n++;
      }
      if (n == 0)
      {
         if (pmx->pcod == 'i')
            err_typ = ERA_OIT;
         else if (pmx->pcod == 'a')
            err_typ = ERA_OOT;
         else if (pmx->pcod == 'l')
            err_typ = ERA_OLT;
      }
      if (n > 1)
      {
         sprintf (errnumb, "%d", n);
         err_txt = errnumb;
         err_typ = pmx->pcod == 'i' ? ERA_MIT : ERA_MOT;
      }
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int   test_markdat (SIM_PTR p_sel)
{
   int   i, j, k;
   MARKDAT_PTR p_mdt;
   LOCUS_PTR p_locus;
   int  *p_a;
   char *p_o;

   NObs = 0;                            /* finds marker data for this        */
   if (!(p_mdt = first_markdat_node))   /* chromosome                        */
   {
      err_typ = ERA_NMD;
      for_chrom (p_sel->chrom_no);
   }
   else
   {
      while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
         p_mdt = p_mdt->p_next;
      if (p_mdt->chrom_no != p_sel->chrom_no)
      {
         err_typ = ERA_NMD;
         for_chrom (p_sel->chrom_no);
      }
      else
      {
         if (p_mdt->no_marks != NMarks) /* checks no. of markers             */
         {
            err_typ = ERA_BND;
            for_chrom (p_mdt->chrom_no);
         }
         else
            NObs = p_mdt->no_obs;
      }
   }
   if (err_typ)
      return FALSE;

   p_o = p_mdt->p_obs;                  /* tests marker alleles within range */
   p_a = p_mdt->p_alls;                 /* ... for each observed individual  */
   for (i = 0; i < NObs; ++i)
   {
      p_locus = first_mark_locus_node;
      for (j = 0; j < NMarks; ++j)      /* ... and each marker               */
      {
         while ((p_locus->p_next && (p_locus->chrom_no < p_sel->chrom_no)
               || (p_locus->chrom_no == p_sel->chrom_no && p_locus->id <= j)))
            p_locus = p_locus->p_next;

         for (k = 0; k < 2; ++k)
         {
            if (*p_a < 0 || *p_a > p_locus->no_alleles)
            {
               sprintf (errnumb, "%s", p_o);
               err_txt = errnumb;
               err_typ = ERA_WMA;
            }
            ++p_a;
         }
      }
      p_o += MAXNAMLEN + 1;
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int   test_probgam (SIM_PTR p_sel)
{
   PROBGAM_PTR p_pg;
   LOCWND_PTR p_node;

   if (!(p_pg = first_probgam_node))    /* expects proband gametes           */
      err_typ = ERA_MPG;

   else if (pmx->pcod == 'i')           /* checks for optional locus window  */
   {                                    /* for ibddrop                       */
      if (p_node = first_locwind_node)
      {                                 /* checks for window for this chrom  */
         while (p_node->p_next && p_node->chrom_no < p_sel->chrom_no)
            p_node = p_node->p_next;

         if (p_node->chrom_no == p_sel->chrom_no)
         {
            LocWndw = p_node->window;
            if (LocWndw < 2 || LocWndw > NLoci)
            {
               err_typ = ERA_LWO;
               sprintf (errnumb, "%d", NLoci);
               err_txt = errnumb;
            }
         }
      }
      if (err_typ)
         ;                              /* tests for error in window         */
      else if (!LocWndw && !(p_pg = first_probgam_node))
         err_typ = ERA_MPG;             /* tests for no window and no gams   */
      else if (!LocWndw && p_pg)
         ;                              /* tests for no window, but gametes  */
      else if (LocWndw && !p_pg)
         err_typ = ERA_MPP;             /* tests for window, but no gametes  */
      else if (LocWndw && p_pg)
         while (p_pg)                   /* for locus window, tests for one   */
         {                              /* pair of gametes in each set       */
            if (p_pg->no_probs != 2)
               err_typ = ERA_PIG;
            p_pg = p_pg->p_next;
         }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int   test_atz_pars (void)
{
   fprintf (stdout, " \n");

   if (!NIt)
   {
      err_typ = ERA_NIT;
      return FALSE;
   }
   if (pmx->pcod == 'i')
   {
      if (err_typ)
         return FALSE;
      else
         return TRUE;
   }

   if (!WIt)
      WIt = NIt;

   if (!FWIt)
      FWIt = NIt;

   if (TraitPheno == UNDEF)
      TraitPheno = N;

   if (SmplSetup == UNDEF)
      SmplSetup = LSMPL;

   if (Scan == UNDEF)
      Scan = N;

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int   test_atz_inpt (void)
{                                       /* tests for enough data in ped rec  */
   /*                                    * allows for gender added when      */
   /*                                    * pedigree checking                 */

   if (pmx->pcod != 'i')
      if (first_tlist_node->id + (SexIn == Y ? 1 : 0) > NInts)
      {
         err_typ = ERA_NEI;
         sprintf (errnumb, "%d", first_tlist_node->id);
         err_txt = errnumb;
         return FALSE;
      }
   return TRUE;
}

int   test_trait_vmap (SIM_PTR p_sim, int *ss_map)
{                                       /* tests for trait map present       */
   /*                                    * (may be gender-specific) for this */
   /*                                    * chromosome & correctly specifiedd */
   int   sx, j;
   VMAP_PTR p_vmap, p_m[3] =
   {NULL, NULL, NULL};

   if (p_sim->no_traits != 1)           /* checks for one trait only         */
   {
      err_typ = ERA_NAT;
      return FALSE;
   }
   /*                                    * checks for sex-specific maps      */
   for (sx = FEM; sx <= AVG; ++sx)      /* present                           */
      if (p_vmap = first_trait_vmap_node[sx])
      {
         while (p_vmap->p_next && p_vmap->chrom_no < p_sim->chrom_no)
            p_vmap = p_vmap->p_next;
         if (p_vmap->chrom_no == p_sim->chrom_no)
         {
            if (sx != AVG)
               *ss_map = TRUE;
            p_m[sx] = p_vmap;
         }
      }
   /*                                    * sets defaults for sex-specific    */
   /*                                    * trait maps not present            */
   /*                                    * checks for at least one map       */
   for (sx = FEM; sx <= AVG; ++sx)      /* present                           */
      if (!p_m[sx])
         if (!(p_m[sx] = p_m[AVG]))
            if (!(p_m[sx] = p_m[FEM]))
               if (!(p_m[sx] = p_m[MAL]))
                  if (p_sim->no_mark > 1)
                     err_typ = ERA_MBM;
                  else
                     err_typ = ERA_MXM;
   if (err_typ)
   {
      for_chrom (p_sim->chrom_no);

      return FALSE;
   }
   /*                                    * checks for sex-specific external  */
   /*                                    * intervals and internal interval   */
   /*                                    * proportions present               */
   for (sx = FEM; sx <= AVG; ++sx)
   {
      if (!p_m[sx]->no_xtrn)
      {
         j = sx;
         if (p_m[AVG]->no_xtrn)
            j = AVG;
         else if (p_m[FEM]->no_xtrn)
            j = FEM;
         else if (p_m[MAL]->no_xtrn)
            j = MAL;
         else
            err_typ = ERA_MXM;

         p_m[sx]->no_xtrn = p_m[j]->no_xtrn;
         p_m[sx]->p_dist = p_m[j]->p_dist;
         p_m[sx]->p_rf = p_m[j]->p_rf;
      }

      if (p_sim->no_mark > 1 && !p_m[sx]->no_prop)
      {
         j = sx;
         if (p_m[AVG]->no_prop)
            j = AVG;
         else if (p_m[FEM]->no_prop)
            j = FEM;
         else if (p_m[MAL]->no_prop)
            j = MAL;
         else
            err_typ = ERA_MIP;

         p_m[sx]->no_prop = p_m[j]->no_prop;
         p_m[sx]->p_prop = p_m[j]->p_prop;
      }
      if (p_m[sx] && p_m[sx]->no_prop && p_sim->no_mark == 1)
      {
         err_typ = ERA_XIP;
         p_m[sx]->no_prop = 0;
      }
   }
   /*                                    * now checks the sex-specific maps  */
   /*                                    * for consistency                   */
   for (sx = FEM; sx < AVG; ++sx)
   {
      if (p_m[sx]->id != p_m[sx + 1]->id)
         err_typ = ERD_TID;
      else if (p_m[sx]->no_xtrn != p_m[sx + 1]->no_xtrn)
         err_typ = ERA_WXD;
      else if (p_m[sx]->no_prop != p_m[sx + 1]->no_prop)
         err_typ = ERA_WIP;
   }
   if (err_typ)
   {
      for_chrom (p_sim->chrom_no);

      return FALSE;
   }
   else
      return TRUE;
}

void  prnt_vmap (VMAP_PTR p_vm, int sx)
{
   int   i;
   char  some[10];

   if (p_vm->no_prop)
   {
      fprintf (stdout, "\n T%d ", p_vm->id);
      switch (sx)
       {
          case MAL:
             fprintf (stdout, "male ");
             break;
          case FEM:
             fprintf (stdout, "female ");
             break;
          default:
             break;
       }
      fprintf (stdout, "map interval proportions between markers:\n");
      for (i = 0; i < p_vm->no_prop; ++i)
      {
         if (!(i % 10))
            fprintf (stdout, "\n");
         sprintf (some, "%5.3f  ", *(p_vm->p_prop + i));
         fprintf (stdout, " %s", &some[1]);
      }
      fprintf (stdout, "\n");
   }

   fprintf (stdout, "\n T%d ", p_vm->id);
   switch (sx)
    {
       case MAL:
          fprintf (stdout, "male ");
          break;
       case FEM:
          fprintf (stdout, "female ");
          break;
       default:
          break;
    }
   fprintf (stdout, "map external distances (cM):\n");

   for (i = 0; i < p_vm->no_xtrn; ++i)
   {
      if (!(i % 10))
         fprintf (stdout, "\n");
      fprintf (stdout, " %4.1f  ", *(p_vm->p_dist + i));
   }
   fprintf (stdout, "\n\n");

   return;
}

void  prnt_vtrait_freqs (SIM_PTR p_sim)
{
   char  prnt_lin[81];

   if (!nxt_trait_vmap (FEM, p_sim))
      return;

   fprintf (stdout, " Trait allele frequencies:\n\n");
   /*                                      * formats the freqs for the lods  */
   /*                                      * trait                           */
   fmt_trait_freqs (prnt_lin, 0, nxt_trait_vmap (FEM, p_sim)->id);

   fprintf (stdout, "%s\n\n", prnt_lin);  /* prints the line                 */
   fflush (stdout);

   return;
}

void  alloc_marek (void)
{
   /* allocs Marek:   2 sexes, no. markers recomb fracts
      /*                 (first recomb fract at [][1]; last one .5)
    */
   int   sx;

   Marek = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, Marek);

   for (sx = FEM; sx <= MAL; ++sx)
   {
      Marek[sx] = malloc ((NMarks + 1) * sizeof (double));
      alloc_chk (WHERE, Marek[sx]);
   }
   return;
}

void  alloc_intval (void)
{
   /* allocates arrays:  IntVal:  no. trait pos + 1
    *                    RecVal:  2 sexes, no. trait pos + 1 recomb fracts  */

   int   sx;

   RecVal = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, RecVal);

   for (sx = FEM; sx <= MAL; ++sx)
   {
      RecVal[sx] = malloc ((NTPos + 1) * sizeof (double));
      alloc_chk (WHERE, RecVal[sx]);
   }
   IntVal = malloc ((NTPos + 1) * sizeof (int));
   alloc_chk (WHERE, IntVal);

   return;
}

void  mak_marek (void)
{
   /* makes Marek[][] : sex-specific recombination fractions
    * (1st and last one set to .5)
    */

   int   sx, entry, i;
   PRNTMAP_PTR p_pm;
   PRNTLOC_PTR p_loc;

   for (sx = FEM; sx <= MAL; ++sx)
   {
      entry = FALSE;
      if (p_pm = nxt_prntmap (&entry, sx))
      {
         p_loc = p_pm->p_locus;

         Marek[sx][0] = .5;             /* ignores trait, if linked and 1st marker */
         for (i = 1; p_loc; p_loc = p_loc->p_next)
         {
            if (p_loc->locus_typ == MD && p_loc->dist != NIL)
               Marek[sx][i++] = dist2rf (p_loc->dist);
         }
         Marek[sx][i] = .5;
      }
   }
   return;
}

void  mak_tarek (void)
{
   /* makes Int0, Rec0[2] - trait position and sex-specific recomb fraction
    *                                          for linked trait              */
   int   sx, entry, i;
   PRNTMAP_PTR p_pm;
   PRNTLOC_PTR p_loc;

   if (NUnlnkd)
   {
      Loc0 = 0;
      Rec0[0] = .5;
      Rec0[1] = .5;
   }
   if (NLnkd)                           /* (a, i)                            */
      for (sx = FEM; sx <= MAL; ++sx)
      {
         entry = FALSE;
         if (p_pm = nxt_prntmap (&entry, sx))
         {
            p_loc = p_pm->p_locus;

            for (i = 0; p_loc; p_loc = p_loc->p_next)
               if (p_loc->locus_typ == TD)
               {
                  Loc0 = p_loc->mark_id;
                  Rec0[sx] = dist2rf (p_loc->dist);
               }
         }
      }
   return;
}

void  mak_intval (SIM_PTR p_sim)
{
   /* makes IntVal[] - list of trial linked-trait positions                  */

   int   i, j;
   VMAP_PTR p_vm;

   p_vm = nxt_trait_vmap (FEM, p_sim);

   for (i = 0; i <= p_vm->no_xtrn; ++i) /* sets loci for extern positions    */
   {                                    /* plus one each end                 */
      IntVal[i] = 0;
      IntVal[NTPos - i] = NMarks;
   }

   IntVal[NTPos] = 0;                   /* and loci for interval proportions */
   for (j = 0; j < (NMarks - 1) * p_vm->no_prop; ++i, ++j)
      IntVal[i] = j / p_vm->no_prop + 1;

   return;
}

void  mak_recval (SIM_PTR p_sim)
{
/* makes RecVal[][] - sex-specific recombination fracts corresp to IntVal[]  */

   VMAP_PTR p_vm;
   PRNTMAP_PTR p_pm;
   PRNTLOC_PTR p_cur;
   int   sx, entry, i, j, k;

   for (sx = FEM; sx <= MAL; ++sx)
   {
      entry = FALSE;
      if ((p_pm = nxt_prntmap (&entry, sx)) &&
          (p_vm = nxt_trait_vmap (sx, p_sim)))
      {
         RecVal[sx][0] = .5;            /* set recomb fracts for each end    */
         RecVal[sx][NTPos] = .5;

         j = p_vm->no_xtrn;             /* sets external recomb fracts       */
         k = NTPos - p_vm->no_xtrn;
         for (i = 0; i < p_vm->no_xtrn; ++i, --j, ++k)
         {
            RecVal[sx][j] = *(p_vm->p_rf + i);
            RecVal[sx][k] = *(p_vm->p_rf + i);
         }

         p_cur = p_pm->p_locus;         /* and recomb fracts between each    */
         while (p_cur && (p_cur = p_cur->p_next))  /* marker pair            */
            for (j = 0; j < p_vm->no_prop; ++j, ++i)
               RecVal[sx][i + 1] =
                  dist2rf (*(p_vm->p_prop + j) * p_cur->loc_dist);
      }
   }
   return;
}

void  mak_allefrq (PRNTMAP_PTR p_pmap, PRNTLOC_PTR p_ploc, int id)
{
   int   j;
   LOCUS_PTR p_locus;

   /*                                      on first entry, mallocs           */
   /*                                         NAll[NLoci] and Qq[NLoci]      */
   /*                                      NAll: 0 .. NLoci-1                */
   /*                                      Qq: 1st subscript:  0 .. NLoci-1  */
   /*                                      Qq: 2nd: 1 .. NAll[1st subscript] */
   /*                                      (Note:  NAll[0], Qq[0] are trait) */
   if (!Qq)
   {
      NAll = malloc (NLoci * sizeof (int));
      alloc_chk (WHERE, NAll);
      Qq = malloc (NLoci * sizeof (double *));
      alloc_chk (WHERE, Qq);
   }

   if (p_ploc && p_ploc->locus_typ == MD)  /* sets no. alleles and stores    */
   {                                    /* allele freqs for marker           */
      p_locus = first_mark_locus_node;
      while ((p_locus->p_next
              && (p_locus->chrom_no < p_pmap->chrom_no)
              || (p_locus->chrom_no == p_pmap->chrom_no
                  && p_locus->id < p_ploc->id)))
         p_locus = p_locus->p_next;

      NAll[id] = p_locus->no_alleles;

      Qq[id] = malloc ((NAll[id] + 1) * sizeof (double));
      alloc_chk (WHERE, Qq[id]);

      Qq[id][0] = 0.;                   /* 1st element is unused             */
      for (j = 0; j < NAll[id]; ++j)
         Qq[id][j + 1] = *(p_locus->p_freq_node->p_freq + j);
   }
   else
   {                                    /* set no. alleles and stores        */
      p_locus = first_trait_locus_node; /* allele freqs for trait            */
      while (p_locus->p_next && p_locus->id < id)
         p_locus = p_locus->p_next;

      NAll[0] = p_locus->no_alleles;

      Qq[0] = malloc ((NAll[0] + 1) * sizeof (double));
      alloc_chk (WHERE, Qq[0]);

      Qq[0][0] = 0.;                    /* 1st element is unused             */
      for (j = 0; j < NAll[0]; ++j)
         Qq[0][j + 1] = *(p_locus->p_freq_node->p_freq + j);
   }
   return;
}

void  mak_markdat (SIM_PTR p_sel)
{
   int   i, j;
   int  *p_a;
   char *p_o;
   MARKDAT_PTR p_mdt;

   if (NObs == 0)
      return;

   if (!(p_mdt = first_markdat_node))   /* finds marker data for this chrom  */
      return;

   while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
      p_mdt = p_mdt->p_next;
   if (p_mdt->chrom_no != p_sel->chrom_no)
      return;

   ObsNam = malloc (NObs * sizeof (char *));  /* mallocs:                    */
   alloc_chk (WHERE, ObsNam);           /*        ObsNam [NObs][MAXNAMLEN+1] */
   ObsIx = malloc (NObs * sizeof (int));    /*      ObsIx  [NObs]              */
   alloc_chk (WHERE, ObsIx);
   ObsMark = malloc (NObs * sizeof (int **));  /* ObsMark[NObs][NMarks][2]   */
   alloc_chk (WHERE, ObsIx);

   for (i = 0; i < NObs; ++i)
   {
      ObsNam[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, ObsNam[i]);
      ObsMark[i] = malloc (NMarks * sizeof (int *));
      alloc_chk (WHERE, ObsMark[i]);
      for (j = 0; j < NMarks; ++j)
      {
         ObsMark[i][j] = malloc (2 * sizeof (int));
         alloc_chk (WHERE, ObsMark[i][j]);
      }
   }

   p_o = p_mdt->p_obs;
   p_a = p_mdt->p_alls;
   for (i = 0; i < NObs; ++i)           /* stores names and alleles          */
   {
      strncpy (ObsNam[i], p_o, MAXNAMLEN + 1);
      p_o += MAXNAMLEN + 1;
      for (j = 0; j < NMarks; ++j)
      {
         ObsMark[i][j][0] = *p_a;
         ++p_a;
         ObsMark[i][j][1] = *p_a;
         ++p_a;
      }
   }
}

void  mak_probgam (void)
{
   int   c, s, i = 0;
   PROBGAM_PTR p_pg;
   PROBLIST_PTR p;

   p_pg = first_probgam_node;
   /*                                    * get and assign last component     */
   while (p_pg->p_next)                 /* with proband gametes              */
      p_pg = p_pg->p_next;

   NScCmp = p_pg->component;

   NScSet = malloc (NScCmp * sizeof (int));
   alloc_chk (WHERE, NScSet);
   NPrGam = malloc (NScCmp * sizeof (int *));
   alloc_chk (WHERE, NPrGam);
   ProbNam = malloc (NScCmp * sizeof (char ***));
   alloc_chk (WHERE, ProbNam);
   ProbIx = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbIx);
   ProbGam = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbGam);
   /*                                    * get and assign no. of proband     */
   p_pg = first_probgam_node;           /* gamete sets for each component    */
   for (c = 0; c < NScCmp; ++c)
   {
      NScSet[c] = 0;
      while (p_pg && p_pg->component <= c + 1)
      {
         ++NScSet[c];
         p_pg = p_pg->p_next;
      }
      if (NScSet[c])
      {
         NPrGam[c] = malloc (NScSet[c] * sizeof (int));
         alloc_chk (WHERE, NPrGam[c]);
         ProbNam[c] = malloc (NScSet[c] * sizeof (char **));
         alloc_chk (WHERE, ProbNam);
         ProbIx[c] = malloc (NScSet[c] * sizeof (int *));
         alloc_chk (WHERE, ProbIx[c]);
         ProbGam[c] = malloc (NScSet[c] * sizeof (int *));
         alloc_chk (WHERE, ProbGam[c]);
      }
      else
      {
         NPrGam[c] = NULL;
         ProbNam[c] = NULL;
         ProbIx[c] = NULL;
         ProbGam[c] = NULL;
      }
   }
   /*                                    * get and assign no. of probands    */
   p_pg = first_probgam_node;           /* for each component set            */
   for (c = 0; c < NScCmp; ++c)
      for (s = 0; s < NScSet[c]; ++s)
         if (NPrGam[c])
         {
            NPrGam[c][s] = p_pg->no_probs;
            if (NPrGam[c][s] > MxGam)
               MxGam = NPrGam[c][s];
            ProbNam[c][s] = malloc (NPrGam[c][s] * sizeof (char *));
            alloc_chk (WHERE, ProbNam[c][s]);
            ProbIx[c][s] = malloc (NPrGam[c][s] * sizeof (int));
            alloc_chk (WHERE, ProbIx[c][s]);
            ProbGam[c][s] = malloc (NPrGam[c][s] * sizeof (int *));
            alloc_chk (WHERE, ProbGam[c][s]);
            p = p_pg->p_pglist;         /* store names and haplotypes of     * 
                                         * proband gametes                   */
            for (i = 0; i < p_pg->no_probs; ++i)
            {
               ProbNam[c][s][i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
               alloc_chk (WHERE, ProbNam[c][s][i]);
               strncpy (ProbNam[c][s][i], p->nam, MAXNAMLEN + 1);
               ProbGam[c][s][i] = p->gamete;
               ++p;
            }
            p_pg = p_pg->p_next;
         }
}

void  mak_atz_inpt (void)
{
   int   j;

   SelInt[0] = 1;                       /* SelInt[]:  data[][] subscripts    */

   if (parentOrdIn == UNDEF || parentOrdIn == PA1ST)
   {
      SelInt[1] = 2;                    /* (father)                          */
      SelInt[2] = 3;                    /* (mother)                          */
   }
   else
   {
      SelInt[1] = 3;                    /* (mother)                          */
      SelInt[2] = 2;                    /* (father)                          */
   }
   SelInt[3] = 4;                       /* (gender)                          */

   /*                                    * NColData is the number of columns
    *                                    * to be allocated to the "data"
    *                                    * array for input_ped()             */
   if (pmx->pcod == 'i')
      NColData = 5;
   else
   {
      for (j = 4; j < 3 + (SexIn == N ? 1 : 0) + NInts; ++j)
      {                                 /* allows for gender added to ped    */
         if (first_tlist_node)
            if (j - 3 == first_tlist_node->id)
               SelInt[j] = 5;           /* the trait is in the ped file      */
            else
               SelInt[j] = FALSE;
      }
      NColData = 6;                     /* names, gender, trait + 1          */
   }
   return;
}

void  prnt_markdat (SIM_PTR p_sel)
{
   int   i, j;

   if (!NObs)
      return;

   if (p_sel->chrom_no)
      fprintf (stdout, " Marker data for chromosome %d:\n\n",
               p_sel->chrom_no);
   else
      fprintf (stdout, " Marker data:\n\n");

   for (i = 0; i < NObs; ++i)
   {
      fprintf (stdout, " %s   ", ObsNam[i]);
      for (j = 0; j < NMarks; ++j)
      {
         if (!((j - 1) % 14) && j != 1)
            fprintf (stdout, "\n        ");
         fprintf (stdout, "%d %d  ", ObsMark[i][j][0], ObsMark[i][j][1]);
      }
      fprintf (stdout, "\n");
   }
   fprintf (stdout, "\n\n");
   fflush (stdout);

   return;
}

void  prnt_probgam (void)
{
   int   c, s, i, n, nc = 0;

   if (LocWndw)
      fprintf (stdout, " Autozygosity will be scored over windows of "
               "%d loci\n\n", LocWndw);

   for (c = 0; c < NScCmp; ++c)
   {
      if (NScSet[c])
      {
         for (s = 0; s < NScSet[c]; ++s)
         {
            for (i = 0; i < NPrGam[c][s]; ++i)
            {
               n = strlen (ProbNam[c][s][i]);
               nc = n > nc ? n : nc;

            }
         }
      }
   }
   fprintf (stdout, " Proband gametes:\n");

   for (c = 0; c < NScCmp; ++c)
   {
      if (NScSet[c])
      {
         fprintf (stdout, "\n Component %d\n\n", c + 1);
         for (s = 0; s < NScSet[c]; ++s)
         {
            fprintf (stdout, " Set %d: ", s + 1);
            for (i = 0; i < NPrGam[c][s]; ++i)
            {
               fprintf (stdout, " %*s", nc, ProbNam[c][s][i]);
               fprintf (stdout, " %d ", ProbGam[c][s][i]);
            }
            fprintf (stdout, "\n");
         }
      }
   }
   fprintf (stdout, "\n");
   fflush (stdout);

   return;
}

void  prnt_atz_pars (void)
{
   if (pmx->pcod == 'i')
   {
      fprintf (stdout, " Number of MC iterations requested:  %d\n\n", NIt);
      return;
   }
   if (TraitPheno)
      fprintf (stdout, " Trait data are phenotypes\n\n");
   else
      fprintf (stdout, " Trait data are genotypes\n\n");

   if (SmplSetup == LSMPL)
      fprintf (stdout, " L-sampler to be used for setup\n\n");
   else if (SmplSetup == SEQIMP)
      fprintf (stdout, " Sequential imputation to be used for setup\n\n");

   fprintf (stdout, " L-sampler probability set to %.3f\n\n", LSamplProb);

   fprintf (stdout, " Number of MC iterations requested:  %d\n\n", NIt);

   if (Scan)
   {
      fprintf (stdout, " Sampling is to be by scan,");
      if (ScIt == 1)
         fprintf (stdout, " with scores estimated every scan\n\n");
      else
         fprintf (stdout, " with scores estimated every %d scans\n\n", ScIt);
   }
   else
   {
      fprintf (stdout, " Sampling is to be by single step,");
      if (ScIt == 1)
         fprintf (stdout, " with scores estimated every step\n\n");
      else
         fprintf (stdout, " with scores estimated every %d steps\n\n", ScIt);
   }
   if (WIt)
      fprintf (stdout, " Progress to be checked every %d iterations\n\n", WIt);
   return;
}
