#include "quant.h"
#include "peelie.h"

/*  Program to peel likelihood for a polygenic trait; version modified 
 *  from TMOLDS/peelie.c program;    additive values only;
 * 
 *  Elizabeth Thompson.  June 1990    
 *
 *  Modifications;  most recent (case SigErr=sigd=0) 12/31/92                 
 *         and corrected for peeling down in this case; 1/9/93
 *         
 *  (SCH) Corrected bug which resulted in i2 occasionally having the
 *  wrong value when peeling down; 10/23/95
 *
 *  (EAT) Will now peel several pedigrees, but does assume that peeling
 *  sequence does each pedigree in turn. Final pivot of each pedigree
 *  should be 0, and the final pivot of final pedigree should be -1.
 *     3/22/97
 *  (peel_quchk modified, as well as peel_quant(), but not so fully checked)
 *
 *  (EAT) Modified to read peeling sequence from integer array; 3/25/97
 *  (MAJ) lik_peel routine transferred to lik_gauss.c; 12/28/98
 *  (EAT) correction to final pivots of pedigrees, now -1 or -2; 5/22/97
 *                    (in peel_quant and peel_quchk)
 *  (MAJ) use global variables SigAdd and SigErr    9/99
 */

static double quadint (double (*matcoeff)[NV], int i, int n);
static double zero_int (double (*matcoeff)[NV], int i, double kk, int n);
static double quadmult (double (*matcoeff)[NV], double *vexh, double c, int n);
static void hvset (double *hv, int it, int id, double kk, double lam, int nn);
static double prevlik (int j, double (*qcoeff)[NV], int pp, double qhk);
static void qcoeffchk (double (*qq)[NV], int nn);
static double periphpar (double sd, double (*qcoeff)[NV],
                         int id, int pp, double kk, double lam, int nn);
static void qchold (int pp, double (*qc)[NV]);
static void evalinit (int *ia, int n);

#ifdef STUFF_COMMENTED_OUT
static void printvec (double *xx, int nn);
#endif /* STUFF_COMMENTED_OUT */

static double quadint (double (*matcoeff)[NV], int i, int n)
/*
 * matrix of coefficients of quadratic form; n variables involved;
 * integration w.r.t. index i is required; returned value is 
 * multiplicative constant 
 */
{
   int   jj, j2, l1, l2;
   double lvec[NV], matset[NV][NV];
   double cc = matcoeff[i][i];
   l1 = -1;
   for (jj = 0; jj < n; ++jj)
   {
      if (jj != i)
      {
         ++l1;
         lvec[l1] = matcoeff[i][jj];
         l2 = -1;
         for (j2 = 0; j2 < n; ++j2)
         {
            if (j2 != i)
            {
               ++l2;
               matset[l1][l2] = matcoeff[jj][j2];
            }
         }
      }
   }
   for (jj = 0; jj < n - 1; ++jj)
   {
      for (j2 = 0; j2 <= jj; ++j2)
      {
         matcoeff[jj][j2] = matset[jj][j2] - lvec[jj] * lvec[j2] / cc;
         matcoeff[j2][jj] = matcoeff[jj][j2];
      }
   }
   for (jj = 0; jj < n; ++jj)
   {
      matcoeff[jj][n - 1] = 0.0;
      matcoeff[n - 1][jj] = 0.0;
   }
   cc = sqrt (PPI / cc);
   return (log (cc));
}

static double zero_int (double (*matcoeff)[NV], int i, double kk, int n)
/*
 * matrix of coefficients of quadratic form; n variables involved;
 * integration w.r.t. index i, in the special case where the variable i 
 * is identically the constant kk  
 */
{
   int   jj, j2, l1, l2;
   double lvec[NV], matset[NV][NV];
   double cc = kk * kk * matcoeff[i][i];
   l1 = -1;
   for (jj = 0; jj < n; ++jj)
   {
      if (jj != i)
      {
         ++l1;
         lvec[l1] = kk * matcoeff[i][jj];
         l2 = -1;
         for (j2 = 0; j2 < n; ++j2)
         {
            if (j2 != i)
            {
               ++l2;
               matset[l1][l2] = matcoeff[jj][j2];
            }
         }
      }
   }
   matset[0][0] += 2.0 * lvec[0] + cc;
   for (jj = 1; jj < n - 1; ++jj)
      matset[jj][0] += lvec[jj];

   for (jj = 0; jj < n - 1; ++jj)
   {
      for (j2 = 0; j2 <= jj; ++j2)
      {
         matcoeff[jj][j2] = matset[jj][j2];
         matcoeff[j2][jj] = matcoeff[jj][j2];
      }
   }
   for (jj = 0; jj < n; ++jj)
   {
      matcoeff[jj][n - 1] = 0.0;
      matcoeff[n - 1][jj] = 0.0;
   }
   return (0.0);
}

static double quadmult (double (*matcoeff)[NV], double *vexh, double c, int n)
/*
 * multiplication of form by added contributions specified by c and vexh 
 * and return of multiplicative constant for density  
 */
{
   int   jj, j2;
   for (jj = 0; jj < n; ++jj)
   {
      for (j2 = 0; j2 < n; ++j2)
         matcoeff[jj][j2] += vexh[jj] * vexh[j2] / c;
   }
   return (-0.5 * log (PPI * c));
}

static void hvset (double *hv, int it, int id, double kk, double lam, int nn)
/*
 * it=(1,4,5),2,3 for data/segregation/popn effect respectively;
 * t = 1 if dominance term requd (e.g. child); it=4 if not (e.g. periphpar) 
 * it = 5 for pivot child (ugh! this is getting horrible). 
 * id = polition of main variable involved 
 * (for it=1, assume dominance in posn 3) 
 * (for it=2 we assume parent values are posns 1 and 2 ); 
 * (for it=5, id=2 and extra 1.0 in posn 1 is reqd.)
 * kk = constant term from data expressions etc.   
 * nn = current number of terms (the const always being the last)  
 */
{
   int   j;
   for (j = 0; j < nn; ++j)
      hv[j] = 0.0;
   hv[0] = kk;
   hv[id] = 1.0;
   if (it == 1)
      hv[3] = 1.0;
   if (it == 2)
   {
      hv[1] = -0.5;
      hv[2] = -0.5;
   }
   if (it == 5)
      hv[1] = 1.0;
   if ((it == 1) || (it == 4) || (it == 5))
      hv[id] *= lam;
}

static double prevlik (int j, double (*qcoeff)[NV], int pp, double qhk)
{
   qcoeff[j][j] += qstore[2][pp];
   qcoeff[j][0] += qstore[1][pp];
   qcoeff[0][j] += qstore[1][pp];
   qcoeff[0][0] += qstore[0][pp];
   return (qhk);
}

static void qcoeffchk (double (*qq)[NV], int nn)
{
   int   i, j;
   printf ("\n qcchk");
   for (i = 0; i < nn; ++i)
   {
      printf ("\n %d", i);
      for (j = 0; j < nn; ++j)
         printf ("  %f", qq[i][j]);
   }
}

static double periphpar (double sd, double (*qcoeff)[NV], int id, int pp,
                         double kk, double lam, int nn)
/*
 * peels a peripheral parent individual, posn id; mean-adj.data kk,
 * number of terms left = nn
 */
{
   double qc, qt, hv[NV];
   qt = 0.0;
   if (istore[pp] != 0)
   {
      istore[pp] = 0;
      /*
       * printf ("\n previous log-likelihood for %s inserted", nghd[pp].self);
       * printf (";  chk prelik %e",kstore[pp]);        
       */
      qt += prevlik (id, qcoeff, pp, kstore[pp]);
      /*
       * printf ("\n after insertion %e  with matrix ", qt);
       * qcoeffchk(qcoeff, nn);
       */
   }
   if (idatu[pp] == 0)
   {
      qc = SigAdd;
      hvset (hv, 3, id, 0.0, 1.0, nn);
      qt += quadmult (qcoeff, hv, qc, nn);
      /* qcoeffchk(qcoeff, nn);   */
   }
   if (idatu[pp] == 2)
      idatu[pp] = 0;

   if (idatu[pp] == 0)
   {
      ++idatu[pp];
      /* printf ("\n inserting data for periphpar %s ", nghd[pp].self);      */
      qc = SigErr + sd;
      if ((qdex[pp] != 0) && (qc > 0))
      {
         hvset (hv, 4, id, -kk, lam, nn);
         qt += quadmult (qcoeff, hv, qc, nn);
      }
   }
   return (qt);
}

static void qchold (int pp, double (*qc)[NV])
{
   qstore[0][pp] = qc[0][0];
   qstore[1][pp] = qc[0][1];
   qstore[2][pp] = qc[1][1];
}

#ifdef STUFF_COMMENTED_OUT
static void printvec (double *xx, int nn)
{
   int   j;
   for (j = 0; j < nn; ++j)
      printf (" %f", xx[j]);
}
#endif /* STUFF_COMMENTED_OUT */

static void evalinit (int *ia, int n)
{
   int   j;
   for (j = 0; j < n; ++j)
      ia[j] = 0;
}

double peel_quant (double sigd, double *y, int *peelseq, double lambda, int im)
/* y is out_fixed data, by position index */
{
   int   nch, npeelch, ipivot, kid, nt, ma, pa, child[NS];
   double qkk, qc, qcoeff[NV][NV], hv[NV], lnlik;
   int   i1, i2, psq, obs, obsm, obsp, domi;
   double lam;

   domi = (sigd == 0.0) ? 0 : 1;
   lnlik = 0.0;
   ipivot = 0;

   if ((SigAdd + sigd) <= 0.0)
   {
      /*
       * printf (" \n WARNING: why polygenic peel with no "
       "additive variance ??");
       * printf (" \n use mjrpeel to peel major gene models");
       * printf (" \n use envonly to compute likelihoods for "
       "non-genetic model"); 
       */

      qkk = envonly (y);

      /* printf (" \n using envonly as you maybe meant to "
         "gives %f",qkk);   */
      return (qkk);
   }
   lam = 1.0;
   if ((im == 1) || (im > 2))
      lam = lambda;

   evalinit (istore, pedsize);
   evalinit (idatu, pedsize);

   psq = 0;
   /* the peeling sequence lists individuals by name */

   /* for each sibship of pedigree in turn */
   while (ipivot > -2)                  /* for final pedigree, 
                                         *    final pivot = -2  */
   {
      nt = 2;

      while (nt > 1)
      {
         for (i1 = 0; i1 < NV; ++i1)
            for (i2 = 0; i2 < NV; ++i2)
               qcoeff[i1][i2] = 0.0;
         qkk = 0.0;
         pa = peelseq[psq];
         ma = peelseq[psq + 1];
         ipivot = peelseq[psq + 2];
         nch = peelseq[psq + 3];

         /*
          * printf ("\n new family pa= %d, ma=%d, pivot=%d\n",pa,ma,ipivot);
          */
         for (i2 = 1; i2 <= nch; ++i2)
            child[i2] = peelseq[psq + 3 + i2];
         psq += 4 + nch;

         /*
          * For the sibship, we shall put the "coeff" variables in the
          * order (const, pat, mat, dom, child ): Now peel each offspring
          * in turn 
          */
         npeelch = 0;
         nt = (domi) ? 4 : 3;
         for (i1 = 1; i1 <= nch; ++i1)
         {
            if (ipivot != child[i1])
            {
               kid = child[i1];

               i2 = (domi) ? 4 : 3;
               /* incorporate segregation on curent child */
               qc = 0.5 * SigAdd;
               ++nt;
               hvset (hv, 2, i2, 0.0, 1.0, nt);
               qkk += quadmult (qcoeff, hv, qc, nt);
               /* incorporate any previous likelihood on child  */
               if (istore[kid] != 0)
               {
                  istore[kid] = 0;
                  qkk += prevlik (i2, qcoeff, kid, kstore[kid]);
               }
               /* incorporate data on current child  */
               obs = qdex[kid];
               ++idatu[kid];
               qc = SigErr + 0.75 * sigd;
               if ((obs != 0) && (qc > 0))
               {
                  hvset (hv, 1, i2, -y[qdex[kid]], lam, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
               }
               /* integrate out current child  */
               if (qc > 0 || obs == 0)
                  qkk += quadint (qcoeff, i2, nt);
               else
                  qkk += zero_int (qcoeff, i2, y[qdex[kid]], nt);
               ++npeelch;
               --nt;

               /*
                * qcoeffchk(qcoeff,nt); 
                * printf("\n child %d peeled; overall likelihood %e\n",
                *                                        child[i1],qkk);
                */
            }
         }

         /* incorporate dominance effects, and peel out if peeling up  */
         if (npeelch == nch)
         {
            if (domi)
            {
               qc = 0.25 * sigd;
               hvset (hv, 3, 3, 0.0, 1.0, nt);
               qkk += quadmult (qcoeff, hv, qc, nt);
               qkk += quadint (qcoeff, 3, nt);
               --nt;
               /*
                * printf ("\n dominance peeled; overall log-likelihood %e".qkk);
                */
            }
            /*
             * incorporate parent data, and additive effects for
             * non-pivot parent   
             */
            /* incorporate any previous likelihood on parent  */
            if (ipivot != ma)
            {
               obs = ma;
               qkk += periphpar (sigd, qcoeff, 2, obs, y[qdex[obs]],
                                 lam, nt);
               /*
                * printf ("\n periph for %d called; log-lhood = %e", ma, qkk);
                */
               if ((qdex[obs] == 0) || (SigErr + sigd > 0))
                  qkk += quadint (qcoeff, 2, nt);
               else
                  qkk += zero_int (qcoeff, 2, y[qdex[obs]], nt);
               --nt;
               /* qcoeffchk(qcoeff,nt);      */
               /* printf ("\n mum peeled; log-lhood= %e",qkk);   */
            }
            if (ipivot != pa)
            {
               obs = pa;
               qkk += periphpar (sigd, qcoeff, 1, obs, y[qdex[obs]],
                                 lam, nt);
               /*
                * printf ("\n periph for %d called; log_lhood = %e", pa, qkk);
                */
               if ((qdex[obs] == 0) || (SigErr + sigd > 0))
                  qkk += quadint (qcoeff, 1, nt);
               else
                  qkk += zero_int (qcoeff, 1, y[qdex[obs]], nt);
               --nt;
               /* qcoeffchk(qcoeff,nt);      */
               /* printf ("\n dad peeled; log-lhood= %e",qkk);  */
            }
         }
         /* All this section above is for peeling up only  */

         /* Following section is for peeling down.  */

         if (npeelch < nch)
         {
            /* peel parents down, including any prev liks  */

            obsm = ma;
            qkk += periphpar (sigd, qcoeff, 2, obsm, y[qdex[obsm]], lam, nt);
            obsp = pa;
            qkk += periphpar (sigd, qcoeff, 1, obsp, y[qdex[obsp]], lam, nt);

            i2 = (domi) ? 4 : 3;

            qc = 0.5 * SigAdd;
            ++nt;
            hvset (hv, 2, i2, 0.0, 1.0, i2 + 1);
            qkk += quadmult (qcoeff, hv, qc, i2 + 1);
            if ((qdex[obsm] == 0) || (SigErr + sigd > 0))
               qkk += quadint (qcoeff, 2, nt);
            else
               qkk += zero_int (qcoeff, 2, y[qdex[obsm]], nt);
            --nt;
            if ((qdex[obsp] == 0) || (SigErr + sigd > 0))
               qkk += quadint (qcoeff, 1, nt);
            else
               qkk += zero_int (qcoeff, 1, y[qdex[obsp]], nt);
            --nt;
            /*
             * printf ("\n peeled parents down; log-lhd = %e with
             * matrix",qkk); qcoeffchk(qcoeff,nt);      
             */

            /*
             * incorporate data on offspring pivot, if any, and peel out
             * dominance. note: if SigErr=sigd=0 postpone putting in this
             * penetrance term until later and make idatu = 2 as an
             * indicator of this fact   
             */

            qc = SigErr + 0.75 * sigd;
            obs = ipivot;

            if (qc > 0)
            {
               ++idatu[obs];
               if (qdex[obs] != 0)
               {
                  hvset (hv, 5, i2 - 2, -y[qdex[obs]], lam, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
               }
               if (domi)
               {
                  qc = 0.25 * sigd;
                  hvset (hv, 3, 1, 0.0, 1.0, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
                  qkk += quadint (qcoeff, 1, nt);
                  --nt;
               }
            }
            else
               idatu[obs] += 2;         /* fix for heritability =1 */
         }
         /* if new pivot has a pre stored up-lik; insert it  */

         if (ipivot > -1)
         {
            kid = ipivot;
            if (istore[kid] != 0)
            {
               istore[kid] = 0;
               qkk += prevlik (1, qcoeff, kid, kstore[kid]);
               /*
                * printf ("\n previous likelihood for %d extracted;",
                * ipivot); printf ("\n log-likelihood = %e  with
                * matrix",qkk); qcoeffchk(qcoeff,nt); 
                */
            }
            /* now restore new likelihoods for this pivot  */
            istore[kid] = 1;
            /* printf ("\n chk new-family store %d %f",kid,qkk);   */
            kstore[kid] = qkk - 0.5 * qcoeff[0][0];
            qcoeff[0][0] = 0.0;
            qchold (kid, qcoeff);
            /* printf (" now storing likelihood for %d;", ipivot);   */
            /* printf ("\n likelihood = %e  with matrix",qkk);   */
            /* qcoeffchk(qcoeff,nt);      */
         }
      }

      /* do final peel of this pedigree to get overall likelihd;   */

      lnlik += (nt == 1) ? (qkk - 0.5 * qcoeff[0][0]) : 0.0;
   }

   /* also check all stored lhoods used  */

   for (i1 = 0; i1 < pedsize; ++i1)
   {
      if (istore[i1] != 0)
         printf ("\n Unused likelihood on %s still stored\n", nghd[i1].self);
      if (idatu[i1] != 1)
         printf ("\n Data on %s used %d times\n", nghd[i1].self, idatu[i1]);
   }

   /* printf("\n ln-likelihood = %e\n", lnlik );    */

   return (lnlik);
}

double peel_quchk (double sigd, double *y, int *peelseq, int lambda, int im)

/* y is out_fixed data, by position index */
{
   int   nch, npeelch, ipivot, kid, nt, ma, pa, child[NS];
   double qkk, qc, qcoeff[NV][NV], hv[NV], lnlik, lam;
   int   i1, i2, psq, obs, obsm, obsp, domi;

   domi = (sigd == 0.0) ? 0 : 1;
   lnlik = 0.0;
   ipivot = 0;

   if ((SigAdd + sigd) <= 0.0)
   {
      give_warn_where (WHERE, ERP_nav, "peel_quchk");
      printf (" \n WARNING: why polygenic peel with no additive variance ??");
      printf (" \n use mjrpeel to peel major gene models");
      printf (" \n use envonly to compute likelihoods for non-genetic model");
      qkk = envonly (y);
      give_warn (ERP_ueo, qkk);
      printf (" \n using envonly as you maybe meant to gives %f", qkk);
      return (qkk);
   }
   lam = 1.0;
   if ((im == 1) || (im > 2))
      lam = lambda;
   evalinit (istore, pedsize);
   evalinit (idatu, pedsize);

   psq = 0;

   /* the peeling sequence lists individuals by name */

   /* for each sibship of the pedigree in turn */

   while (ipivot > -2)                  /* for final pedigree, 
                                         *    final pivot = -2  */
   {
      nt = 2;

      while (nt > 1)
      {
         for (i1 = 0; i1 < NV; ++i1)
            for (i2 = 0; i2 < NV; ++i2)
               qcoeff[i1][i2] = 0.0;
         qkk = 0.0;
         pa = peelseq[psq];
         ma = peelseq[psq + 1];
         ipivot = peelseq[psq + 2];
         nch = peelseq[psq + 3];

         /*
          * printf ("\n new family pa= %d, ma=%d, pivot=%d\n", pa, ma, ipivot);
          */
         for (i2 = 1; i2 <= nch; ++i2)
            child[i2] = peelseq[psq + 3 + i2];
         psq += 4 + nch;

         /*
          * For the sibship, we shall put the "coeff" variables in the
          * order (const, pat, mat, dom, child ): Now peel each offspring
          * in turn 
          */
         npeelch = 0;
         nt = (domi) ? 4 : 3;
         for (i1 = 1; i1 <= nch; ++i1)
         {
            if (ipivot != child[i1])
            {
               kid = child[i1];

               i2 = (domi) ? 4 : 3;
               /* incorporate segregation on curent child */
               qc = 0.5 * SigAdd;
               ++nt;
               hvset (hv, 2, i2, 0.0, 1.0, nt);
               qkk += quadmult (qcoeff, hv, qc, nt);
               /* incorporate any previous likelihood on child  */
               if (istore[kid] != 0)
               {
                  istore[kid] = 0;
                  qkk += prevlik (i2, qcoeff, kid, kstore[kid]);
               }
               /* incorporate data on current child  */
               obs = qdex[kid];
               ++idatu[kid];
               qc = SigErr + 0.75 * sigd;
               if ((obs != 0) && (qc > 0))
               {
                  hvset (hv, 1, i2, -y[qdex[kid]], lam, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
               }
               /* integrate out current child  */
               if (qc > 0 || obs == 0)
                  qkk += quadint (qcoeff, i2, nt);
               else
                  qkk += zero_int (qcoeff, i2, y[qdex[kid]], nt);
               ++npeelch;
               --nt;

               qcoeffchk (qcoeff, nt);
               printf ("\n child %d peeled; overall log-likelihood %e\n",
                       child[i1], qkk);
            }
         }

         /* incorporate dominance effects, and peel out if peeling up  */
         if (npeelch == nch)
         {
            if (domi)
            {
               qc = 0.25 * sigd;
               hvset (hv, 3, 3, 0.0, 1.0, nt);
               qkk += quadmult (qcoeff, hv, qc, nt);
               qkk += quadint (qcoeff, 3, nt);
               --nt;
               printf ("\n dominance peeled; overall log-likelihood %e", qkk);
            }
            /*
             * incorporate parent data, and additive effects for 
             * non-pivot parent   
             */
            /* incorporate any previous likelihood on parent  */
            if (ipivot != ma)
            {
               obs = ma;
               qkk += periphpar (sigd, qcoeff, 2, obs, y[qdex[obs]],
                                 lam, nt);
               printf ("\n periph for %d called; log-lhood= %e", ma, qkk);
               if ((qdex[obs] == 0) || (SigErr + sigd > 0))
                  qkk += quadint (qcoeff, 2, nt);
               else
                  qkk += zero_int (qcoeff, 2, y[qdex[obs]], nt);
               --nt;
               qcoeffchk (qcoeff, nt);
               printf ("\n mum peeled; log-lhood= %e", qkk);
            }
            if (ipivot != pa)
            {
               obs = pa;
               qkk += periphpar (sigd, qcoeff, 1, obs, y[qdex[obs]],
                                 lam, nt);
               printf ("\n periph for %d called; log-lhood= %e", pa, qkk);
               if ((qdex[obs] == 0) || (SigErr + sigd > 0))
                  qkk += quadint (qcoeff, 1, nt);
               else
                  qkk += zero_int (qcoeff, 1, y[qdex[obs]], nt);
               --nt;
               qcoeffchk (qcoeff, nt);
               printf ("\n dad peeled; log-lhood= %e", qkk);
            }
         }
         /* All this section above is for peeling up only  */

         /* Following section is for peeling down.  */

         if (npeelch < nch)
         {

            /* peel parents down, including any prev liks  */

            obsm = ma;
            qkk += periphpar (sigd, qcoeff, 2, obsm, y[qdex[obsm]], lam, nt);
            qcoeffchk (qcoeff, nt);
            obsp = pa;
            qkk += periphpar (sigd, qcoeff, 1, obsp, y[qdex[obsp]], lam, nt);
            qcoeffchk (qcoeff, nt);
            qc = 0.5 * SigAdd;
            ++nt;
            hvset (hv, 2, i2, 0.0, 1.0, i2 + 1);
            qkk += quadmult (qcoeff, hv, qc, i2 + 1);
            qcoeffchk (qcoeff, nt);
            if ((qdex[obsm] == 0) || (SigErr + sigd > 0))
               qkk += quadint (qcoeff, 2, nt);
            else
               qkk += zero_int (qcoeff, 2, y[qdex[obsm]], nt);
            --nt;
            qcoeffchk (qcoeff, nt);
            if ((qdex[obsp] == 0) || (SigErr + sigd > 0))
               qkk += quadint (qcoeff, 1, nt);
            else
               qkk += zero_int (qcoeff, 1, y[qdex[obsp]], nt);
            --nt;
            qcoeffchk (qcoeff, nt);
            printf ("\n peeled parents down; log-lhd = %e with matrix", qkk);
            qcoeffchk (qcoeff, nt);

            /*
             * incorporate data on offspring pivot, if any, and peel out
             * dominance. note: if SigErr=sigd=0 postpone putting in this
             * penetrance term until later and make idatu = 2 as an
             * indicator of this fact   
             */

            qc = SigErr + 0.75 * sigd;
            obs = ipivot;

            if (qc > 0)
            {
               ++idatu[obs];
               if (qdex[obs] != 0)
               {
                  hvset (hv, 5, i2 - 2, -y[qdex[obs]], lam, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
               }
               if (domi)
               {
                  qc = 0.25 * sigd;
                  hvset (hv, 3, 1, 0.0, 1.0, nt);
                  qkk += quadmult (qcoeff, hv, qc, nt);
                  qkk += quadint (qcoeff, 1, nt);
                  --nt;
               }
            }
            else
               idatu[obs] += 2;         /* fix for heritability =1 */
         }
         /* if new pivot has a pre stored up-lik; insert it  */

         if (ipivot > -1)
         {
            kid = ipivot;
            if (istore[kid] != 0)
            {
               istore[kid] = 0;
               qkk += prevlik (1, qcoeff, kid, kstore[kid]);
               printf ("\n previous likelihood for %d extracted;", ipivot);
               printf ("\n like,ihood = %e  with matrix", qkk);
               qcoeffchk (qcoeff, nt);
            }
            /* now restore new likelihoods for this pivot  */
            istore[kid] = 1;
            kstore[kid] = qkk - 0.5 * qcoeff[0][0];
            qcoeff[0][0] = 0.0;
            qchold (kid, qcoeff);
            printf (" now storing likelihood for %d;", ipivot);
            printf ("\n log-likelihood = %e  with matrix", qkk);
            qcoeffchk (qcoeff, nt);
         }
      }
      /* do final peel to get overall likelihd for this component */
      lnlik += (nt == 1) ? (qkk - 0.5 * qcoeff[0][0]) : 0.0;
   }
   /* also check all stored lhoods used  */
   for (i1 = 0; i1 < pedsize; ++i1)
   {
      if (istore[i1] != 0)
         printf ("\n unused lhd on %d still stored", data[i1][1]);
      if (idatu[i1] != 1)
         printf ("\n data on %d used %d times", data[i1][1], idatu[i1]);
   }
   /* printf("\n ln-likelihood = %e\n", lnlik );     */

   return (lnlik);
}
