#include "quant.h"

/***************************************************************************
   Implemented June 1990 to allow 3 alternative methods of likelihood
   evaluation -- direct computation, the Thompson and Shaw EM formula,  
   and polygenic peeling.
  
       Routine envonly, for independent normals, added 10/4/1991.
  
       Routine lik_big (Thompson & Shaw univar likelihood for big pedigrees)
						  added 9/2/1993
  
       Peeling sequence specification in lik_peel modified, March 1997, EAT.

       Routine lik_peel removed to another file, to separate peeling stuff
            (so peelie.h no longer needs to be included)    , May 1997, EAT.
  
***************************************************************************/

double matprod (double *xx, double **pp, int n)
{
   int   i, j;
   double cc = 0.0;
   for (i = 1; i <= n; ++i)
   {
      for (j = 1; j <= n; ++j)
         cc += xx[i] * xx[j] * pp[i][j];
   }
   return (cc);
}

double invtrc (double *ev)
{
   int   j;
   double cc = 0.0;
   for (j = 1; j <= np; ++j)
      cc += 1.0 / (SigErr + SigAdd * ev[j]);
   return (cc);
}

double logdet (double *ev)
{
   int   j;
   double cc = 0.0;
   for (j = 1; j <= np; ++j)
      cc += log (SigErr + SigAdd * ev[j]);
   return (cc);
}

double invdet (double *ev)          /* here ev are eigens of inverse */
{
   int   j;
   double cc = 0.0;
   for (j = 1; j <= np; ++j)
      cc += log (SigErr + SigAdd / ev[j]);
   return (cc);
}

void  likval (int k, double *evgen, double *y, double *a, double *epsi,
              double **genmat)

/* original EM equation version of likelihood evaluation  */
{
   double lhood;

   lhood = -0.5 * np * log (PPI);
   lhood -= 0.5 * logdet (evgen);
   lhood -= 0.5 * (innerprod (y, y, np) - innerprod (a, a, np)) / SigErr;
   lhood += 0.5 * matprod (epsi, genmat, np) * SigAdd / (SigErr * SigErr);

   poly_print_pars (k, 1);
   printf ("\n current log-likelihood, a la Thompson & Shaw    =");
   print_val (lhood);
   printf ("\n");
   fflush (stdout);
}

void  lik_big (int k, double *evgen, double *y, double *a, double *b, Term * ginv, int nt)

/* original EM equation version of likelihood evaluation  */
{
   double lhood;

   lhood = -0.5 * np * log (PPI);
   lhood -= 0.5 * invdet (evgen);
   lhood -= 0.5 * (innerprod (y, y, np) - innerprod (a, a, np)) / SigErr;
   lhood += 0.5 * quarz (b, b, ginv, nt) / SigAdd;

   poly_print_pars (k, 2);
   printf ("\n current log-likelihood, a la Thompson & Shaw =");
   print_val (lhood);
   printf ("\n\n\n");
   fflush (stdout);
}

void  lik_direct (double *evgen, double *y, double **genmat)
/* direct likelihood evaluation by inversion of variance matrix */

{
   double lhood, **vvv;
   int   i, j;

   vvv = malloc ((np + 1) * sizeof (double *));
   alloc_chk (WHERE, vvv);
   vvv[0] = malloc ((np + 1) * (np + 1) * sizeof (double));
   alloc_chk (WHERE, vvv[0]);
   for (i = 1; i <= np; ++i)
      vvv[i] = vvv[i - 1] + np + 1;

   lhood = -0.5 * np * log (PPI);
   lhood -= 0.5 * logdet (evgen);
   for (i = 1; i <= np; ++i)
      for (j = i; j <= np; ++j)
      {
         vvv[i][j] = 0.0;
         if (i == j)
            vvv[i][j] = SigErr;
         vvv[i][j] += SigAdd * genmat[i][j];
         vvv[j][i] = vvv[i][j];
      }
   invert_mat (vvv, np);
   lhood -= 0.5 * matprod (y, vvv, np);

   printf (" current log-likelihood, brute force computation =");
   print_val (lhood);
   printf ("\n");
   fflush (stdout);

   free ((void *) vvv[0]);
   vvv[0] = NULL;
   free ((void *) vvv);
   vvv = NULL;
}

/*************************************************************************
  
  purpose: calculate the log likelihood for independent normals
      e.g. model with environmental component only,  or
	  for given major-gene combination, when not additive variance.
  
  ln f ( y | G ) = - k/2 * log ( 2 * Pi * sig_env ) - 1 / ( 2 * sig_env )
                   * sum_i xx_i^2
  
  where xx_i = y_i - mu(geno);
  
  date: 2/13/91; updated and installed in quant 10/4/91.
  
**************************************************************************/

double envonly (double *xx)
/**  xx is the mean-adjusted data-vector, SigErr the variance   **/

{
   double cc = 0.0;

   cc = np * log (PPI * SigErr) + totvar (xx) / SigErr;

   cc = -0.5 * cc;
   return (cc);
}

/**************************************************************************
  
  date 5/25/92; routine totvar() added, 
                to permit basepoint ln-lhd computation.            (EAT)

***************************************************************************/

double totvar (double *xx)
{
   int   i;
   double cc = 0.0;

   for (i = 1; i <= np; ++i)
      cc += xx[i] * xx[i];
   return (cc);
}
